/*
 * Copyright (c) 2016 Samsung Electronics Co., Ltd
 *
 * Licensed under the Flora License, Version 1.1 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://floralicense.org/license/
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <Elementary.h>
#include <efl_extension.h>
#include "$(appName).h"
#include "view.h"
#include "view_defines.h"
#include "thread.h"

typedef enum { BT_START, BT_STOP, BT_LAP, BT_RESET } bt_state_e;
typedef enum { BT_LEFT, BT_RIGHT } bt_type_e;

static void _time_update_cb(long hour, long min, long sec, long msec);
static int _stop(void);

typedef struct _lap_data {
	int count;
	int l_op;
	long lhour;
	long lmin;
	long lsec;
	long lmsec;
} lap_data;

static struct view_info {
	Evas_Object *win;
	Evas_Object *conform;
	Evas_Object *layout_display;
	Evas_Object *box;
	Evas_Object *list;
	Evas_Object *btn_start;
	Evas_Object *btn_lap;
	Evas_Object *btn_stop;
	Evas_Object *btn_reset;
	Elm_Genlist_Item_Class itc_full;
	Eina_Bool start;
	int count;
	int l_op;
	int desc;
	thread_info_s th_info;
	Ecore_Thread *thread;
	Eina_Bool stopped;
	bt_state_e bt_state_left;
	bt_state_e bt_state_right;
	char edj_path[PATH_MAX];
} s_info = {
	.win = NULL,
	.conform = NULL,
	.layout_display = NULL,
	.box = NULL,
	.list = NULL,
	.btn_start = NULL,
	.btn_lap = NULL,
	.btn_stop = NULL,
	.btn_reset = NULL,
	.itc_full = {0,},
	.start = EINA_FALSE,
	.count = 0,
	.l_op = 155,
	.desc = 1,
	.th_info = {
		.lhour = 0,
		.lmin = 0,
		.lsec = 0,
		.lmsec = 0,
		.startTime = 0.0,
		.elapsedTime = 0.0,
		.time_update_func = _time_update_cb,
		.time_stop_func = _stop,
	},
	.thread = NULL,
	.stopped = EINA_FALSE,
	.bt_state_left = BT_START,
	.bt_state_right = BT_LAP,
	.edj_path = {0,},
};

static void _delete_win_request_cb(void *data, Evas_Object *obj, void *event_info);
static void _layout_back_cb(void *data, Evas_Object *obj, void *event_info);
static Evas_Object *_on_content_get_cb(void *data, Evas_Object *obj, const char *part);
static void _button_clicked_cb(void *data, Evas_Object *obj, const char *emission, const char *source);
static void _time_update_cb(long hour, long min, long sec, long msec);
static void _get_app_resource(const char *edj_file_in, char *edj_path_out);
static Eina_Bool _create_main_layout(void);
static Eina_Bool _create_display(Evas_Object *parent);
static Eina_Bool _create_buttons(Evas_Object *parent);
static void _format_time(long time, const char *part_name1, const char *part_name2);
static void _display_time(long hour, long min, long sec, long msec);
static void _start(void);
static void _lap(void);
static void _reset(void);

/**
 * @brief Creates essential objects: window, conformant and layout.
 * @param[in] user_data The user data.
 * @return This function returns EINA_TRUE on successful view creation,
 * otherwise EINA_FALSE is returned.
 */
Eina_Bool view_create(void *user_data)
{
	/* Create the window */
	s_info.win = view_create_win(PACKAGE);
	if (s_info.win == NULL) {
		dlog_print(DLOG_ERROR, LOG_TAG, "failed to create a window.");
		return EINA_FALSE;
	}

	/* Create the conformant */
	s_info.conform = view_create_conformant(s_info.win);
	if (s_info.conform == NULL) {
		dlog_print(DLOG_ERROR, LOG_TAG, "failed to create a conformant");
		return EINA_FALSE;
	}

	if (!_create_main_layout())
		return EINA_FALSE;

	/* Show the window after main view is set up */
	evas_object_show(s_info.win);

	return EINA_TRUE;
}

/**
 * @brief Creates a basic window named package.
 * @param[in] pkg_name Name of the window.
 * @return This function returns window object if created successfully,
 * otherwise NULL is returned.
 */
Evas_Object *view_create_win(const char *pkg_name)
{
	Evas_Object *win = NULL;

	/*
	 * Window
	 * Create and initialize elm_win.
	 * elm_win is mandatory to manipulate the window.
	 */
	win = elm_win_util_standard_add(pkg_name, pkg_name);
	elm_win_conformant_set(win, EINA_TRUE);
	elm_win_indicator_mode_set(win, ELM_WIN_INDICATOR_SHOW);
	elm_win_indicator_opacity_set(win, ELM_WIN_INDICATOR_BG_TRANSPARENT);
	elm_win_autodel_set(win, EINA_TRUE);

	evas_object_smart_callback_add(win, "delete,request", _delete_win_request_cb, NULL);

	return win;
}

/**
 * @brief Creates a layout object and sets it to the parent object.
 * The resulting layout is based on the provided EDJE script.
 * @param[in] parent The object to which the layout is to be set.
 * @param[in] file_path File path of EDJ file.
 * @param[in] group_name Name of a group from EDJE file to be used.
 * newly created layout is to be set.
 * @return This function returns layout object if created successfully,
 * otherwise NULL is returned.
 */
Evas_Object *view_create_layout(Evas_Object *parent, const char *file_path, const char *group_name)
{
	Evas_Object *layout = NULL;

	/* Create layout using EDC(an edje file) */
	layout = elm_layout_add(parent);
	if (!layout) {
		dlog_print(DLOG_ERROR, LOG_TAG, "elm_layout_add() failed.");
		return NULL;
	}

	if (!elm_layout_file_set(layout, file_path, group_name)) {
		dlog_print(DLOG_ERROR, LOG_TAG, "elm_layout_file_set() failed.");
		return NULL;
	}

	/* Layout size setting */
	evas_object_size_hint_weight_set(layout, EVAS_HINT_EXPAND, EVAS_HINT_EXPAND);
	eext_object_event_callback_add(layout, EEXT_CALLBACK_BACK, _layout_back_cb, NULL);

	evas_object_show(layout);

	return layout;
}

/**
 * @brief Creates a conformant object.
 * @param[in] win The window object where the conformant is to be set.
 * Conformant is mandatory for base GUI to have proper size.
 * @return This function returns conformant object if created successfully,
 * otherwise NULL is returned.
 */
Evas_Object *view_create_conformant(Evas_Object *win)
{
	/*
	 * Conformant
	 * Create and initialize elm_conformant.
	 * elm_conformant is mandatory for base GUI to have proper size
	 * when indicator or virtual keypad is visible.
	 */
	Evas_Object *conform = NULL;

	if (win == NULL) {
		dlog_print(DLOG_ERROR, LOG_TAG, "window is NULL.");
		return NULL;
	}

	conform = elm_conformant_add(win);
	evas_object_size_hint_weight_set(conform, EVAS_HINT_EXPAND, EVAS_HINT_EXPAND);
	elm_win_resize_object_add(win, conform);
	elm_object_signal_emit(conform, "elm,state,indicator,overlap", "elm");
	evas_object_show(conform);

	return conform;
}

/**
 * @brief Creates a button object.
 * @param[in] parent The parent object where the button is to be set.
 * @param[in] part_name The name of the parent's part were the button is to be set.
 * @param[in] icon The path to the icon file to be displayed as a button.
 * @param[in] on_click_cb The click event callback function handler.
 * @param[in] data The data to be passed to the callback function.
 * @return This function returns button object if created successfully,
 * otherwise NULL is returned.
 */
Evas_Object *view_create_button(Evas_Object *parent, const char *target_part_name, const char *icon, Edje_Signal_Cb on_click_cb, void *data)
{
	Evas_Object *button = elm_layout_add(parent);
	if (!button) {
		dlog_print(DLOG_ERROR, LOG_TAG, "elm_layout_add() failed.");
		return NULL;
	}

	evas_object_size_hint_weight_set(button, EVAS_HINT_EXPAND, EVAS_HINT_EXPAND);

	if (!elm_layout_file_set(button, s_info.edj_path, icon)) {
		dlog_print(DLOG_ERROR, LOG_TAG, "elm_layout_file_set() failed.");
		return NULL;
	}

	elm_layout_signal_callback_add(button, "elm,action,click", "*", on_click_cb, data);

	return button;
}

/**
 * @brief Internal callback function invoked when a genlist item is going to be deleted.
 * @param[in] data User data.
 * @param[in] obj The genlist.
 */
static void _item_del_cb(void *data, Evas_Object *obj)
{
	if (data)
		free(data);
}

/**
 * @brief Creates a list object.
 * @param[in] parent The parent object where the list is to be set.
 * @return This function returns list object if created successfully,
 * otherwise NULL is returned.
 */
Evas_Object *view_create_list(Evas_Object *parent)
{
	Evas_Object *genlist = NULL;

	s_info.itc_full.item_style = "full";
	s_info.itc_full.func.text_get = NULL;
	s_info.itc_full.func.content_get = _on_content_get_cb;
	s_info.itc_full.func.state_get = NULL;
	s_info.itc_full.func.del = _item_del_cb;

	genlist = elm_genlist_add(parent);
	if (!genlist) {
		dlog_print(DLOG_ERROR, LOG_TAG, "elm_scroller_add() failed.");
		return NULL;
	}

	evas_object_size_hint_weight_set(genlist, EVAS_HINT_EXPAND,	EVAS_HINT_EXPAND);
	evas_object_size_hint_align_set(genlist, EVAS_HINT_FILL, EVAS_HINT_FILL);
	elm_genlist_mode_set(genlist, ELM_LIST_COMPRESS);
	elm_genlist_homogeneous_set(genlist, EINA_TRUE);
	elm_scroller_policy_set(genlist, ELM_SCROLLER_POLICY_OFF, ELM_SCROLLER_POLICY_OFF);
	elm_object_part_content_set(parent, PART_MAIN_LIST, genlist);
	evas_object_show(genlist);

	return genlist;
}

/**
 * @brief Destroys window and frees its resources.
 */
void view_destroy(void)
{
	if (s_info.win == NULL)
		return;

	ecore_thread_cancel(s_info.thread);
	evas_object_del(s_info.win);
}

/**
 * @brief Internal callback function invoked when the main window needs to be destroyed.
 * @param[in] data The user data passed to the evas_object_smart_callback_add() function.
 * @param[in] obj The object invoking this callback function.
 * @param[in] event_info The structure containing the information on this event.
 */
static void _delete_win_request_cb(void *data, Evas_Object *obj, void *event_info)
{
	ui_app_exit();
}

/**
 * @brief Internal callback function invoked on HW Back button press.
 * @param[in] data The user data passed to the eext_object_event_callback_add() function.
 * @param[in] obj The object invoking this callback function.
 * @param[in] event_info The structure containing the information on this event.
 */
static void _layout_back_cb(void *data, Evas_Object *obj, void *event_info)
{
	/* Let window go to hide state. */
	elm_win_lower(s_info.win);
}

/**
 * @brief Internal callback function invoked on list's item display.
 * This function is assigned to the func.content_get handler of genlist item's class structure.
 * It is responsible for items layout creation, text part population with provided text content
 * and background color assignment to create gradient effect across all appended items.
 * @param[in] data The user data passed to the elm_genlist_item_append() function.
 * @param[in] obj The object invoking this callback function.
 * @param[in] event_info The structure containing the information on this event.
 * @return The layout object to be displayed in item's content area.
 */
static Evas_Object *_on_content_get_cb(void *data, Evas_Object *obj, const char *part)
{
	char buf[PATH_MAX] = {0,};
	Evas_Object *bg = NULL;
	Evas_Object *item = NULL;
	lap_data *_data = (lap_data *)data;

	if (!data) {
		dlog_print(DLOG_ERROR, LOG_TAG, "data() is NULL.");
		return NULL;
	}

	item = view_create_layout(s_info.list, s_info.edj_path, GRP_LIST_ITEM);
	if (!item) {
		dlog_print(DLOG_ERROR, LOG_TAG, "view_create_layout() failed.");
		return NULL;
	}

	snprintf(buf, PATH_MAX, "Lap %02d", _data->count);
	elm_object_part_text_set(item, PART_LIST_ITEM_LAP, buf);
	snprintf(buf, PATH_MAX, "%02ld:%02ld:%02ld.%02ld", _data->lhour, _data->lmin, _data->lsec, _data->lmsec);
	elm_object_part_text_set(item, PART_LIST_ITEM_LAP_TIME, buf);

	bg =  elm_bg_add(item);
	if (!bg) {
		dlog_print(DLOG_ERROR, LOG_TAG, "elm_bg_add() failed.");
		return NULL;
	}

	elm_object_part_content_set(item, PART_LIST_ITEM_BG, bg);
	evas_object_color_set(bg, 170, 170, 170, _data->l_op);

	return item;
}

/**
 * @brief Internal callback function invoked on button click.
 * This function handles all the action buttons (START, STOP, RESET, LAP).
 * Depending on its current state, on click the relevant action is executed.
 * @param[in] data The user data passed to the view_create_button() function.
 * @param[in] obj The object invoking this callback function.
 * @param[in] event_info The structure containing the information on this event.
 */
static void _button_clicked_cb(void *data, Evas_Object *obj, const char *emission, const char *source)
{
	bt_type_e type = (bt_type_e)data;

	if (type == BT_LEFT) {
		if (s_info.bt_state_left == BT_START)
			_start();
		else if (s_info.bt_state_left == BT_STOP)
			_stop();
		else
			dlog_print(DLOG_ERROR, LOG_TAG, "Not handled button state.");
	} else if (type == BT_RIGHT) {
		if (s_info.bt_state_right == BT_LAP)
			_lap();
		else if (s_info.bt_state_right == BT_RESET)
			_reset();
		else
			dlog_print(DLOG_ERROR, LOG_TAG, "Not handled button state.");
	} else {
		dlog_print(DLOG_ERROR, LOG_TAG, "Not handled button type.");
	}
}

/**
 * @brief Internal callback function invoked when the time elapses.
 * This function updates the time displayed.
 * @param[in] hour The current hour to be displayed.
 * @param[in] min The current minute to be displayed.
 * @param[in] sec The current second to be displayed.
 * @param[in] msec The current millisecond to be displayed.
 */
static void _time_update_cb(long hour, long min, long sec, long msec)
{
	_display_time(hour, min, sec, msec);
}

/**
 * @brief Internal function which creates fully qualified path to the provided resource file.
 * @param[in] edj_file_in The file name.
 * @param[out] edj_path_out The fully qualified path to the edj_file_in file.
 */
static void _get_app_resource(const char *edj_file_in, char *edj_path_out)
{
	char *res_path = app_get_resource_path();
	if (res_path) {
		snprintf(edj_path_out, PATH_MAX, "%s%s", res_path, edj_file_in);
		free(res_path);
	}
}

/**
 * @brief Internal function which creates the main layout.
 * @return This function returns EINA_TRUE on successful layout creation,
 * otherwise EINA_FALSE is returned.
 */
static Eina_Bool _create_main_layout(void)
{
	Evas_Object *layout;

	_get_app_resource(EDJ_MAIN, s_info.edj_path);
	layout = view_create_layout(s_info.conform, s_info.edj_path, GRP_MAIN);
	if (!layout)
		return EINA_FALSE;

	elm_object_content_set(s_info.conform, layout);
	s_info.layout_display = layout;

	if (!_create_display(layout))
		return EINA_FALSE;

	if (!_create_buttons(layout))
		return EINA_FALSE;

	s_info.list = view_create_list(layout);
	if (!s_info.list)
		return EINA_FALSE;

	evas_object_show(layout);

	return EINA_TRUE;
}

/**
 * @brief Internal function which creates the display view.
 * @param[in] parent The parent object for display.
 * @return This function returns EINA_TRUE on successful display creation,
 * otherwise EINA_FALSE is returned.
 */
static Eina_Bool _create_display(Evas_Object *parent)
{
	elm_object_part_text_set(parent, PART_MAIN_SEMI1, ":");
	elm_object_part_text_set(parent, PART_MAIN_SEMI2, ":");

	_display_time(0, 0, 0, 0);

	return EINA_TRUE;
}

/**
 * @brief Internal function which creates the action buttons.
 * @param[in] The parent object for buttons.
 * @return This function returns EINA_TRUE on successful action buttons creation,
 * otherwise EINA_FALSE is returned.
 */
static Eina_Bool _create_buttons(Evas_Object *parent)
{
	s_info.btn_stop = view_create_button(s_info.layout_display,
				PART_MAIN_BUTTONS_LEFT, GRP_BT_STOP, _button_clicked_cb,
				(void *)BT_LEFT);
	if (!s_info.btn_stop)
		return EINA_FALSE;

	s_info.btn_start = view_create_button(s_info.layout_display,
			PART_MAIN_BUTTONS_LEFT, GRP_BT_START, _button_clicked_cb,
			(void *)BT_LEFT);
	if (!s_info.btn_start)
		return EINA_FALSE;

	elm_object_part_content_set(s_info.layout_display, PART_MAIN_BUTTONS_LEFT, s_info.btn_start);
	evas_object_show(s_info.btn_start);

	s_info.btn_reset = view_create_button(s_info.layout_display,
			PART_MAIN_BUTTONS_RIGHT, GRP_BT_RESET, _button_clicked_cb,
				(void *)BT_RIGHT);
	if (!s_info.btn_reset)
		return EINA_FALSE;

	s_info.btn_lap = view_create_button(s_info.layout_display,
			PART_MAIN_BUTTONS_RIGHT, GRP_BT_LAP, _button_clicked_cb,
			(void *)BT_RIGHT);
	if (!s_info.btn_lap)
		return EINA_FALSE;

	elm_object_part_content_set(s_info.layout_display, PART_MAIN_BUTTONS_RIGHT, s_info.btn_reset);
	evas_object_show(s_info.btn_reset);

	return EINA_TRUE;
}

/**
 * @brief Internal function which formats given time component and puts it to the display.
 * @param[in] time The time component to be formatted and displayed.
 * @param[in] part_name1 The part representing the first digit of time component.
 * @param[in] part_name2 The part representing the second digit of time component.
 */
static void _format_time(long time, const char *part_name1, const char *part_name2)
{
	char buff[256] = {0,};
	long tmp = 0;

	if (time >= 10) {
		tmp = time % 10;
		if (tmp != 0) {
			snprintf(buff, sizeof(buff), "%01ld", tmp);
			elm_object_part_text_set(s_info.layout_display, part_name2, buff);
		} else {
			snprintf(buff, sizeof(buff), "%01ld", tmp);
			elm_object_part_text_set(s_info.layout_display, part_name2, buff);
			tmp = time / 10;
			snprintf(buff, sizeof(buff), "%01ld", tmp);
			elm_object_part_text_set(s_info.layout_display, part_name1, buff);
		}
	} else {
		tmp = 0;
		snprintf(buff, sizeof(buff), "%01ld", tmp);
		elm_object_part_text_set(s_info.layout_display, part_name1, buff);
		snprintf(buff, sizeof(buff), "%01ld", time);
		elm_object_part_text_set(s_info.layout_display, part_name2, buff);
	}
}

/**
 * @brief Internal function which displays the current time.
 * @param[in] hour The current hour to be displayed.
 * @param[in] min The current minute to be displayed.
 * @param[in] sec The current second to be displayed.
 * @param[in] msec The current millisecond to be displayed.
 */
static void _display_time(long hour, long min, long sec, long msec)
{
	char buff[256] = {0,};

	_format_time(hour, PART_MAIN_H1, PART_MAIN_H2);
	_format_time(min, PART_MAIN_M1, PART_MAIN_M2);
	_format_time(sec, PART_MAIN_S1, PART_MAIN_S2);

	/* msec */
	snprintf(buff, sizeof(buff), ".%02ld", msec);
	elm_object_part_text_set(s_info.layout_display, PART_MAIN_MSEC, buff);
}

/**
 * @brief Internal function which changes the action button's image.
 * @param[in] button The target button object.
 * @param[in] part_name The name of the part occupied by the target button.
 * @param[in] button2 The new button to be set.
 */
static void _change_icon(Evas_Object *button, const char *part_name, Evas_Object *button2)
{
	evas_object_hide(button);
	elm_object_part_content_unset(s_info.layout_display, part_name);
	elm_object_part_content_set(s_info.layout_display, part_name, button2);
	evas_object_show(button2);
}

/**
 * @brief Internal function which starts the stopwatch.
 */
static void _start(void)
{
	_change_icon(s_info.btn_start, PART_MAIN_BUTTONS_LEFT, s_info.btn_stop);
	_change_icon(s_info.btn_reset, PART_MAIN_BUTTONS_RIGHT, s_info.btn_lap);

	s_info.bt_state_left = BT_STOP;
	s_info.bt_state_right = BT_LAP;
	s_info.start = EINA_TRUE;
	s_info.th_info.lhour = 0;
	s_info.th_info.lmin = 0;
	s_info.th_info.lsec = 0;
	s_info.th_info.lmsec = 0;

	/* run stopwatch */
	if (s_info.stopped)
		s_info.th_info.startTime = ecore_time_get() - s_info.th_info.elapsedTime;
	else
		s_info.th_info.startTime = ecore_time_get();

	/* tread run */
	s_info.thread = ecore_thread_feedback_run(thread_job, thread_job_notify,
			thread_job_end, NULL, &s_info.th_info, EINA_FALSE);

	if (!s_info.thread)
		dlog_print(DLOG_ERROR, LOG_TAG, "ecore_thread_feedback_run() failed.");
}

/**
 * @brief Internal function which stops the stopwatch.
 */
static int _stop(void)
{
	_change_icon(s_info.btn_stop, PART_MAIN_BUTTONS_LEFT, s_info.btn_start);
	_change_icon(s_info.btn_lap, PART_MAIN_BUTTONS_RIGHT, s_info.btn_reset);

	s_info.bt_state_left = BT_START;
	s_info.bt_state_right = BT_RESET;
	s_info.stopped = EINA_TRUE;

	ecore_thread_cancel(s_info.thread);
	s_info.thread = NULL;

	return TIZEN_ERROR_NONE;
}

/**
 * @brief Internal function which catches the lap time while the stopwatch is running.
 */
static void _lap(void)
{
	Evas_Object *list = NULL;
	Elm_Object_Item *item = NULL;
	lap_data *data = NULL;

	if (!s_info.start)
		return;

	list = s_info.list;

	s_info.count++;
	if (s_info.desc > 0) {
		if (s_info.l_op < 255)
			s_info.l_op += 25;
		else {
			s_info.desc = 0;
			s_info.l_op -= 25;
		}
	} else {
		if (s_info.l_op > 155)
			s_info.l_op -= 25;
		else {
			s_info.desc = 1;
			s_info.l_op += 25;
		}
	}

	data = (lap_data *)calloc(1, sizeof(lap_data));
	if (!data) {
		dlog_print(DLOG_ERROR, LOG_TAG, "lap_data allocation failed.");
		return;
	}

	data->count = s_info.count;
	data->l_op = s_info.l_op;
	data->lhour = s_info.th_info.lhour;
	data->lmin = s_info.th_info.lmin;
	data->lsec = s_info.th_info.lsec ;
	data->lmsec = s_info.th_info.lmsec;

	item = elm_genlist_item_append(list, &(s_info.itc_full), data, NULL, ELM_GENLIST_ITEM_NONE, NULL, NULL);
	if (!item) {
		dlog_print(DLOG_ERROR, LOG_TAG, "elm_genlist_item_append() failed.");
		return;
	}

	elm_genlist_item_bring_in(item, ELM_GENLIST_ITEM_SCROLLTO_IN);
}

/**
 * @brief Internal function which resets the stopwatch.
 */
static void _reset(void)
{
	s_info.bt_state_left = BT_START;
	s_info.bt_state_right = BT_LAP;
	s_info.th_info.lhour = 0;
	s_info.th_info.lmin = 0;
	s_info.th_info.lsec = 0;
	s_info.th_info.lmsec = 0;
	s_info.count = 0;
	s_info.l_op = 155;
	s_info.desc = 1;
	s_info.start = EINA_FALSE;
	s_info.stopped = EINA_FALSE;

	elm_genlist_clear(s_info.list);

	_display_time(0, 0, 0, 0);
}
